/*
 * Copyright 2008 Berend "Kirk" Wouda
 * 
 * CheezeFix is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * CheezeFix is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with CheezeFix; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


// Cheezefixer.
package com.kirkwarez.cheezefix;

// Import the IO and regular expression classes.
import java.io.*;
import java.util.*;
import java.util.regex.*;
 
/**
 * This class can fix wiremod MDL files by cheeze, which fixes the console error spam.
 * 
 * @author Berend "Kirk" Wouda
 * @version 1.00
 * @since 1.00
 */
public class CheezeFix {
	/**
	 * Constructs a new <code>CheezeFix</code>.
	 * 
	 * @param backup Whether or not to back up files before they are fixed.
	 */
	public CheezeFix(boolean backup) {
		// Save the backup option.
		this.backup = backup;
		
		// Initialize.
		fixedMDLCount = 0;
		skippedMDLCount = 0;
		skippedNonMDLCount = 0;
	}
	
	
	/**
	 * Fixes the given file or the contents of the given directory (recursively).
	 * 
	 * @param file The file to fix, or the directory to fix the contents of.
	 */
	public void fix(File file) {
		// Make sure the file exists.
		if(file.exists()) {
			// Check type.
			if(file.isDirectory()) {
				// Log.
				System.out.println("Entering '" + file.getPath() + "'...");
				
				// Go into directories.
				fixDir(file);
			}
			else if(file.isFile()) {
				// Fix only MDL files.
				if(file.getName().endsWith(".mdl")) {
					// Log.
					System.out.println("Handling '" + file.getPath() + "'...");
					
					// Fix the file.
					fixFile(file);
				}
				else {
					// Skip it.
					System.out.println("Skipping '" + file.getPath() + "'... (File is not an MDL file.)");

					// Count it.
					skippedNonMDLCount++;
				}
			}
			else {
				// Uhm...
				throw new Error("Seems like the programmer done an' fucked up.");
			}
		}
		else {
			// Skip it.
			System.out.println("Skipping '" + file.getPath() + "'... (File does not exist.)");

			// Count it.
			skippedNonExistingCount++;
		}
	}
	
	/**
	 * Fixes the contents of the given directory (recursively).
	 * 
	 * @param dir The directory to fix the contents of.
	 */
	protected void fixDir(File dir) {
		// Go through all the files in the dir.
		for(File file : dir.listFiles()) {
			// Fix the file.
			fix(file);
		}
	}
	
	/**
	 * Fixes the given file.
	 * 
	 * @param file The file to fix.
	 */
	protected void fixFile(File file) {
		// Catch IO errors.
		try {
			// Log.
			System.out.println("- Reading '" + file.getPath() + "'...");
			
			// Read in the file.
			byte[] brokenData = readFile(file);
	
			// Log.
			System.out.println("- Fixing '" + file.getPath() + "'...");
			
			// Check and optionally change the data.
			byte[] fixedData = fixData(brokenData);
			
			// Check the data.
			if(fixedData != null) {
				// Check if we should make a backup.
				if(backup) {
					// Log.
					System.out.println("- Creating backup for '" + file.getPath() + "'...");
					
					// Create a backup.
					writeFile(new File(file.getPath() + ".bak"), brokenData);
				}
				
				// Log.
				System.out.println("- Writing '" + file.getPath() + "'...");
				
				// Write out the file.
				writeFile(file, fixedData);
				
				// Count it.
				fixedMDLCount++;
			}
			else {
				// Skip it.
				System.out.println("- Skipping '" + file.getPath() + "'... (File is not broken.)");

				// Count it.
				skippedMDLCount++;
			}
		}
		catch(IOException e) {
			// Skip it.
			System.out.println("- Skipping '" + file.getPath() + "'... (IO error ocured.)");

			// Count it.
			skippedMDLCount++;
		}
	}
	
	
	/**
	 * Reads a file into an array.
	 * 
	 * @param file The file to read.
	 * @return The read data.
	 * @throws IOException If a reading error occurs.
	 */
	protected byte[] readFile(File file) throws IOException {
		// Open up a byte stream.
		DataInputStream input = new DataInputStream(new BufferedInputStream(new FileInputStream(file)));
		
		// The data array (if the file is larger than 2 GB this will just screw up :P).
		byte[] data = new byte[(int) file.length()];
		
		// Read the entire contents of the file into the data array.
		input.readFully(data);
		
		// Close the file.
		input.close();
		
		// Return the data array.
		return data;
	}
	
	/**
	 * Writes an array into a file.
	 * 
	 * @param file The file to write to.
	 * @param data The data to write.
	 * @throws IOException If a writing error occurs.
	 */
	protected void writeFile(File file, byte[] data) throws IOException {
		// Open up a byte stream.
		BufferedOutputStream output = new BufferedOutputStream(new FileOutputStream(file));
		
		// Write the data array to the file.
		output.write(data);
		
		// Close the file.
		output.close();
	}
	
	
	/**
	 * Inspects the given data, and fixes it if possible.
	 * 
	 * @param data The data to fix.
	 * @return The fixed data, or <code>null</code> if the data is fine or could not be
	 * fixed.
	 */
	protected byte[] fixData(byte[] data) {
		// Catch errors...
		try {
			// Copy the data array.
			data = data.clone();
			
			// Convert the data into an ASCII string.
			String file = new String(data, "ASCII");
			
			// Search for treasure.
			int mkvIndex = file.indexOf("mdlkeyvalue");
			
			// Check whether we found treasure.
			if(mkvIndex >= 0) {
				// Find the string's null terminator.
				int ntIndex = file.indexOf(0x00, mkvIndex);
				
				// Check whether we found it.
				if(ntIndex >= 0) {
					// Take out the mdlkeyvalue section.
					String mkvSection = file.substring(mkvIndex, ntIndex);
					
					// Create a matcher for it.
					Matcher mkvMatcher = cheezePattern.matcher(mkvSection);
					
					// Check whether we can fix it.
					if(mkvMatcher.matches()) {
						// We can fix it. Create the new mdlkeyvalue section.
						String fixedMKVSection = "mdlkeyvalue{\nprop_data{\"base\" \"" + mkvMatcher.group(1) + "\"}\n}";
						
						// Convert it into bytes.
						byte[] fixedMKVSectionData = fixedMKVSection.getBytes("ASCII");
						
						// Cram it into the data.
						System.arraycopy(fixedMKVSectionData, 0, data, mkvIndex, fixedMKVSectionData.length);
						
						// Return the data.
						return data;
					}
					else {
						// The data is not broken, or at least not in a way we can fix.
						return null;
					}
				}
				else {
					// The data is broken in another way.
					return null;
				}
			}
			else {
				// The data does not contain an mdlkeyvalue section, so there's nothing
				// broken for us to fix.
				return null;
			}
		}
		catch(UnsupportedEncodingException e) {
			// Uhm...
			throw new Error("Seems like the programmer done an' fucked up.");
		}
	}
	
	
	/**
	 * Returns the fixed MDL files count.
	 * 
	 * @return The fixed MDL files count.
	 */
	public int getFixedMDLCount() {
		// Return the count.
		return fixedMDLCount;
	}
	
	
	/**
	 * Returns the skipped MDL files count.
	 * 
	 * @return The skipped MDL files count.
	 */
	public int getSkippedMDLCount() {
		// Return the count.
		return skippedMDLCount;
	}
	
	
	/**
	 * Returns the skipped non-MDL files count.
	 * 
	 * @return The skipped non-MDL files count.
	 */
	public int getSkippedNonMDLCount() {
		// Return the count.
		return skippedNonMDLCount;
	}
	
	
	/**
	 * Returns the skipped non-existing files count.
	 * 
	 * @return The skipped non-existing files count.
	 */
	public int getSkippedNonExistingCount() {
		// Return the count.
		return skippedNonExistingCount;
	}
	
	
	/**
	 * Whether or not to back up files before they are fixed.
	 */
	protected boolean backup;
	
	/**
	 * The amount of MDL files fixed.
	 */
	protected int fixedMDLCount;
	
	/**
	 * The amount of MDL files skipped.
	 */
	protected int skippedMDLCount;
	
	/**
	 * The amount of non-MDL files skipped.
	 */
	protected int skippedNonMDLCount;
	
	/**
	 * The amount of non-existing files skipped.
	 */
	protected int skippedNonExistingCount;
	
	
	
	/**
	 * The pattern of the mdlkeyvalue section.
	 */
	public static Pattern mkvPattern = Pattern.compile("mdlkeyvalue\\p{Space}*\\{.*?prop_data\\p{Space}*\\{\\p{Space}*(?:\"[^\"]*\"\\p{Space}*\"[^\"]*\"\\p{Space}*)*\\}.*?\\}\\p{Space}", Pattern.DOTALL);
	
	/**
	 * The pattern of the mdlkeyvalue section in cheeze's MDL files.
	 */
	public static Pattern cheezePattern = Pattern.compile("mdlkeyvalue\\n\\{\\nprop_data \\{\"base\" ([^ ]+) \\}\\n");
	
	
	
	/**
	 * Fixes the given wiremod MDL files by cheeze, which fixes the console error spam.
	 * 
	 * @param args The commandline parameters, which should be files and directories.
	 */
	public static void main(String[] args) {
		// Yay :P
		System.out.println();
		System.out.println("CheezeFix v1.00 by Kirk");
		System.out.println();
		
		// Commandline options.
		boolean backup = false;
		
		// Commandline files and directories.
		LinkedList<File> list = new LinkedList<File>();
		
		// Go through the commandline parameters.
		for(String parameter : args) {
			// Check if it's one of the possible options.
			if(parameter.equals("--backup") || parameter.equals("-b"))  backup = true;
			else {
				// It's a file or directory. Add it to the list.
				list.add(new File(parameter));
			}
		}
		
		// Check if we got anything.
		if(list.size() == 0) {
			// Usage :P
			System.out.println("Usages:");
			System.out.println("- CheezeFix.bat [option] <directory or file> <directory or file> ...");
			System.out.println("- java -jar CheezeFix.jar [option] <directory or file> <directory or file> ...");
			System.out.println();
			System.out.println("Commandline options:");
			System.out.println("- --backup or -b");
			System.out.println("  Creates a backup of every file before it is fixed.");
			System.out.println();
		}
		else {
			// Catch errors.
			try {
				// Create our fixer.
				CheezeFix fix = new CheezeFix(backup);
				
				// Fix the given files and directories.
				for(File file : list) {
					// Fix the file or directory.
					fix.fix(file);
				}
				
				// Show results.
				System.out.println();
				System.out.println("MDL files fixed: " + fix.getFixedMDLCount());
				System.out.println("MDL files skipped: " + fix.getSkippedMDLCount());
				System.out.println("Non-MDL files skipped: " + fix.getSkippedNonMDLCount());
				System.out.println("Non-existing files skipped: " + fix.getSkippedNonExistingCount());
				System.out.println();
			}
			catch(Error e) {
				// Noes!
				System.out.println("A fatal error occured:");
				e.printStackTrace(System.out);
				System.out.println();
			}
		}
	}
}